function [errors]=fem_2d_errors(uh,model,uex,uexx,uexy,varargin)
% FEM_2D_ERRORS Computes errors for 2d b.v.p. solved by FEM-P1 or FEM-P2
% [errors]=fem_2d_errors(uh,model,uex,uexx,uexy)
% [errors]=fem_2d_errors(uh,model,uex,uexx,uexy,errtype)
% [errors]=fem_2d_errors(uh,model,uex,uexx,uexy,errtype,degree)
%
% Input:
%        uh= numerical solution
%        model = model object generated by MATLAB (it must contain the
%        field model.Mesh)
%          uex =  function handle @(x) with the exact solution
%          uexx =  function handle @(x) with the 1st-prder x-derivative 
%                  of the exact solution
%          uexy =  function handle @(x) with the 1st-prder y-derivative 
%                  of the exact solution
%optional input: errtype =  0 : absolute errors (default value)
%                        =  1 : relative errors
%                degree = degree of exactness of quadrature formulas
%                         (default value =4), possible values: 1, 3, 4, 10
% Output: 
% 
%         errors.inf = ||u-uh|| with respect to maximum-norm
%         errors.h1 = ||u-uh|| with respect to H1-norm
%         errors.l2 = ||u-uh|| with respect to L2-norm
%

%   Written by Paola Gervasio
%   $Date: 2020/05/07$


degree=4;
 param=zeros(10,1);
 
 if nargin==6
   param(8)=varargin{1};
   degree=4;
 elseif nargin==7
   param(8)=varargin{1};
   degree=varargin{2};
   D=[1 3 4 10];
   if isempty(intersect(D,degree))
     fprintf('degree = %d is not available, instead it was set the default value degree = 4\n',degree);
     degree=4;
   end
end
   
fun_set.uex=uex;
fun_set.uexx=uexx;
fun_set.uexy=uexy;

 [p,e,t]=meshToPet(model.Mesh);
vmesh.nodes=p';
vmesh.femtype=2;
if strcmp(model.Mesh.GeometricOrder,'linear')
vmesh.p=1;
else
vmesh.p=2;
end
 vmesh.nov=t(1:vmesh.p*3,:);
 vmesh.ne=size(t,2);

[basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree) ;
[errors]=errors_fem_2d(uh,vmesh,basis,jacf,qf,fun_set,param);
end
% -----------------------------------------------------
function [basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree)
% GROUND_ZERO_FEM_2d: functions, jacobians, normal versors
% this function works on all the elements of the domain
%  [basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree,qtype) 
% Input: vmesh = structure with the mesh on the domain
%              .femtype =2 P
%              .p = local polynomial degree (1 or 2)
%              .nodes (nodes of the mesh)
%              .nov (connectivity matrix of the mesh)
%              .ne =  number of elements
%        degree: precision degree of quadrature formulas 
% Output: 
%         basis: structure with fields
%          .phi phi_j(x_q) Lagrange basis at the quadrature nodes
%          .dphix dphi_j/dx(x_q) dx of the Lagrange basis at quad nodes
%          .dphiy dphi_j/dy(x_q) dy of the Lagrange basis at quad nodes
%         jac: cell 
%         jacf{ie}, with ie=1,...ne structure with fields:
%                 .jac = det(J_F_ie) determinant of the Jacobian of the map
%                 .Jmt(2,2) = J^{-T} matrix (used to build the stiffness
%                 matrix) (it is costant on the element)
%                 .J(2,2) = J_F_ie matrix (used to build the stiffness
%                 matrix) (it is costant on the element)
%         qf = structure with fields
%          .x, .w quadrature nodes and weights
% (optional output) boundary =  structure with information on the boundary
%          NOTE:  if this output is required, then the function
%                 boundary_fem_2d.m is called and vmesh.edges is needed
% 
%   Written by Paola Gervasio
%   $Date: 2019/04/11$


% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% compute quadrature nodes and weights in the 2d reference simplex
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


[qf]=quadrature_simplex_2d(degree);



% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% basis functions and its derivatives at the quadrature nodes in the
% reference simplex
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if vmesh.p==1 % P1
    [basis]=basis_p1_2d(qf);
    elseif vmesh.p==2 % P2
    [basis]=basis_p2_2d(qf);
    end    


% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% basis functions and derivatives at the quadrature nodes in the
% Jacobian matrices
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
jacf=cell(vmesh.ne,1);

for ie=1:vmesh.ne
nov_l=vmesh.nov(:,ie);
jacf{ie}=cosjac_simplex_2d(vmesh.nodes(nov_l,:));
end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% compute quadrature nodes and weights in each element of the mesh
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
[nq,~]=size(qf.xq);
ne=vmesh.ne;
qf.xq_ph=zeros(nq,ne);
qf.yq_ph=zeros(nq,ne);
qf.wq_ph=zeros(nq,ne);


for ie=1:vmesh.ne
  novl=vmesh.nov(:,ie);
  V=vmesh.nodes(novl(1),1:2);
  B=jacf{ie}.J;
  jac=jacf{ie}.jac;
X=[qf.xq';qf.yq'];

X1=B*X+[V(1)*ones(1,nq);V(2)*ones(1,nq)];
xq=X1(1,:)';
yq=X1(2,:)';

qf.xq_ph(:,ie)=xq;
qf.yq_ph(:,ie)=yq;
qf.wq_ph(:,ie)=qf.wq*jac;

end
end


% -----------------------------------------------------
function [qf]=quadrature_simplex_2d(degree);
% quadrature_simplex_2d : quadrature nodes and weights in the 2d simplex
% [qf]=quadrature_simplex_2d(degree);
% Input: degree= precision degree of the quadrature formula
% Ouput: qf.xq, .yq, .zq (coordinates of the nodes)
%        qf.wq weights

if degree==1
wq=1/6*ones(3,1);
xq=[0,1,0]';
yq=[0,0,1]';

elseif degree==3
% degree= 3
wq=1/2*[-0.5625;ones(3,1)*25/48];
xq=[1/3,0.2,0.6,0.2]';
yq=[1/3,0.2,0.2,0.6]';

elseif degree==4
% degree= 4
wq =1/2*[0.223381589678011.*ones(1,3) 0.109951743655322.*ones(1,3)]';
xq = [0.445948490915965,0.108103018168070,0.445948490915965,...
     0.091576213509771,0.816847572980459,0.091576213509771]';
yq = [0.445948490915965,0.445948490915965,0.108103018168070,...
     0.091576213509771,0.091576213509771,0.816847572980459]';

elseif degree==10
% degree= 10
wq = 1/2*[0.090817990382754 0.036725957756467*ones(1,3) 0.045321059435528*ones(1,3) 0.072757916845420*ones(1,6) ...
         0.028327242531057*ones(1,6) 0.009421666963733*ones(1,6)]';
    xq = [1/3,0.485577633383657,0.028844733232685,0.485577633383657,...
       0.109481575485037,0.781036849029926,0.109481575485037,0.307939838764121,...
       0.550352941820999,0.141707219414880,0.550352941820999,0.141707219414880,...
       0.307939838764121,0.246672560639903,0.728323904597411,0.025003534762686,...
       0.728323904597411,0.025003534762686,0.246672560639903,0.066803251012200,...
       0.923655933587500,0.009540815400299,0.923655933587500,0.009540815400299,...
       0.066803251012200]';
    yq = [1/3,0.485577633383657,0.485577633383657,0.028844733232685,...
       0.109481575485037,0.109481575485037,0.781036849029926,0.550352941820999,...
       0.307939838764121,0.550352941820999,0.141707219414880,0.307939838764121,...
       0.141707219414880,0.728323904597411,0.246672560639903,0.728323904597411,...
       0.025003534762686,0.246672560639903,0.025003534762686,0.923655933587500,...
       0.066803251012200,0.923655933587500,0.009540815400299,0.066803251012200,...
       0.009540815400299]';
end
qf.xq=xq;
qf.wq=wq;
qf.yq=yq;
end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [basis]=basis_p1_2d(qf)
% basis_p1_2d: computes quadrature nodes and weights and evaluate there 
% the P1-basis on the reference triangle T with vertices
%  (0,0,0) (1,0,0), (0,1,0) 
% [basis]=basis_p1_2d(degree)
%
% Input: qf structure with quadrature nodes
% Output:
%        basis.phi = Lagrange basis functions evaluated at the quadrature nodes 
%        basis.phix, .phiy  = Lagrange basis functions derivatives 
%                        evaluated at the quadrature nodes 
%        basis.phi_wq = array of values:  phi_i(x_q) * w_q
%        basis.phi_ij_wq = array of values: phi_j(x_q) * phi_i(x_q) * w_q
%        basis.mat00_{ij}=(phi_j,phi_i)_T
%        basis.mat11_{ij}=(dphi_j/dx,dphi_i/dx)_T
%        basis.mat12_{ij}=(dphi_j/dx,dphi_i/dy)_T
%        basis.mat12_{ij}=(dphi_j/dx,dphi_i/dz)_T
%        basis.mat22_{ij}=(dphi_j/dy,dphi_i/dy)_T

%   Written by Paola Gervasio
%   $Date: 2015/04/15$
%   $Date: 2019/04/27$

nq=length(qf.xq);
zz=zeros(nq,1); % vettore nullo
uno=ones(nq,1);

% basis functions 
basis.phi=[1-qf.xq-qf.yq, qf.xq, qf.yq];
[~,nb]=size(basis.phi);
% derivates:
% dphi/dx
basis.phix=[-uno,uno,zz];
% dphi/dy
basis.phiy=[-uno,zz,uno];

% phi *wq
basis.phi_wq=zeros(nq,nb);
for i=1:nb
basis.phi_wq(:,i)=basis.phi(:,i).*qf.wq;
end

% phi *wq
basis.phi_ij_wq=zeros(nq,nb,nb);

for i=1:nb
for j=1:nb
basis.phi_ij_wq(:,j,i)=basis.phi(:,i).*basis.phi_wq(:,j);
end
end

% mat00_{ij}=(phi_j,phi_i)_T
basis.mat00=[ 1, 1/2, 1/2; 
              1/2, 1, 1/2 ; 
              1/2, 1/2, 1 ]/12;
% mat1_{ij}=dphi_j/dx_i
basis.mat1=[ ]; 

 

  % mat11_{ij}=(dphi_j/dx,dphi_i/dx)_T
basis.mat11=0.5*[1,-1,0; -1,1,0; 0,0,0];

% mat12_{ij}=(dphi_j/dx,dphi_i/dy)_T
basis.mat12=0.5*[1,-1,0; 0,0,0;-1,1,0];

% mat22_{ij}=(dphi_j/dy,dphi_i/dy)_T
basis.mat22=0.5*[1,0,-1; 0,0,0; -1,0,1];
end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [basis]=basis_p2_2d(qf)
% basis_p2_2d: computes quadrature nodes and weights and evaluate there 
% the P2-basis on the reference triangle T with vertices
%  (0,0,0) (1,0,0), (0,1,0)
% [basis]=basis_p2_2d(degree)
%
% Input: qf structure with quadrature nodes
% Output:
%        basis.phi = Lagrange basis functions evaluated at the quadrature nodes 
%        basis.phix, .phiy = Lagrange basis functions derivatives 
%                        evaluated at the quadrature nodes 
%        basis.phi_wq = array of values:  phi_i(x_q) * w_q
%        basis.phi_ij_wq = array of values: phi_j(x_q) * phi_i(x_q) * w_q
%        basis.mat00_{ij}=(phi_j,phi_i)_T
%        basis.mat11_{ij}=(dphi_j/dx,dphi_i/dx)_T
%        basis.mat12_{ij}=(dphi_j/dx,dphi_i/dy)_T
%        basis.mat22_{ij}=(dphi_j/dy,dphi_i/dy)_T

%   Written by Paola Gervasio
%   $Date: 2015/04/15$
%   $Date: 2019/04/27$

nq=length(qf.xq);
zz=zeros(nq,1); % vettore nullo

xq=qf.xq; yq=qf.yq;

% basis functions 
xyq=xq.*yq;
xq2=xq.^2;
yq2=yq.^2;


basis.phi=[1-3*xq-3*yq+4*xyq+2*xq2+2*yq2,...
  -xq+2*xq2, -yq+2*yq2, 4*xq-4*xyq-4*xq2, 4*xyq, 4*yq-4*xyq-4*yq2];
%
% derivates:
% dphi/dx
basis.phix=[-3+4*yq+4*xq, -1+4*xq, zz, 4-4*yq-8*xq, 4*yq, -4*yq];
%
% dphi/dy
basis.phiy=[-3+4*xq+4*yq, zz, -1+4*yq, -4*xq ,4*xq, 4-4*xq-8*yq];

[~,nb]=size(basis.phi);


% phi *wq
basis.phi_wq=zeros(nq,nb);
for i=1:nb
basis.phi_wq(:,i)=basis.phi(:,i).*qf.wq;
end

% phi *wq
basis.phi_ij_wq=zeros(nq,nb,nb);

for i=1:nb
for j=1:nb
basis.phi_ij_wq(:,j,i)=basis.phi(:,i).*basis.phi_wq(:,j);
end
end

% mat00_{ij}=(phi_j,phi_i)_T
basis.mat00=[ 6, -1, -1,  0, -4,  0;
  -1,  6, -1,  0,  0, -4;
  -1, -1,  6, -4,  0,  0;
   0,  0, -4, 32, 16, 16;
  -4,  0,  0, 16, 32, 16;
   0, -4,  0, 16, 16, 32]/360;

% mat1_{ij}=dphi_j/dx_i
basis.mat1= [];
 
  % mat11_{ij}=(dphi_j/dx,dphi_i/dx)_T
basis.mat11=[  3,  1, 0, -4,  0,  0;
   1,  3, 0, -4,  0,  0;
   0,  0, 0,  0,  0,  0;
  -4, -4, 0,  8,  0,  0;
   0,  0, 0,  0,  8, -8;
   0,  0, 0,  0, -8,  8]/6;


 % mat12_{ij}=(dphi_j/dx,dphi_i/dy)_T
basis.mat12= [  3,  1, 0, -4,  0,  0;
   0,  0, 0,  0,  0,  0;
   1, -1, 0,  0,  4, -4;
   0, -4, 0,  4, -4,  4;
   0,  4, 0, -4,  4, -4;
  -4,  0, 0,  4, -4,  4]/6;



% mat22_{ij}=(dphi_j/dy,dphi_i/dy)_T
basis.mat22=[  3, 0,  1,  0,  0, -4;
   0, 0,  0,  0,  0,  0;
   1, 0,  3,  0,  0, -4;
   0, 0,  0,  8, -8,  0;
   0, 0,  0, -8,  8,  0;
  -4, 0, -4,  0,  0,  8]/6;

% ~~~~~~~~~~~~~~~~~~
% only for Stokes P2-P1
% mat10_{ij}=(dphi_j/dx,\eta_i)_T (per stokes P2-P1)

basis.mat10=[ -1, 0, 0,  1, 1, -1;
           0, 1, 0, -1, 1, -1;
           0, 0, 0,  0, 2, -2]/6;

% mat20_{ij}=(dphi_j/dy,\eta_i)_T (per stokes P2-P1)

basis.mat20=[ -1, 0, 0, -1, 1,  1;
           0, 0, 0, -2, 2,  0;
           0, 0, 1, -1, 1, -1]/6;
end
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [jacf]=cosjac_simplex_2d(nodes)
% COSJAC_SIMPLEX_2d: evaluation of the Jacobian of the map on one element
%
% [jacf]=cosjac_simplex_2d(nodes)
% Input: basis = structure with Lagrange basis functions derivatives at the
%                 LGL nodes
%        nodes = mesh nodes in the element 
% Output: jacf.jac = determinant of J_F
%             .Jmt(2,2) = J^{-T} matrix =J^cof/det(J) 
%             .J(2,2) = J_F matrix 

%   Written by Paola Gervasio
%   $Date: 2019/04/27
 
x=nodes(:,1); y=nodes(:,2);  

jacf.J=[x(2)-x(1), x(3)-x(1);
y(2)-y(1),y(3)-y(1)];
jacf.jac=det(jacf.J);
jacf.Jmt=(jacf.J')\eye(2);

end
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [errors]=errors_fem_2d(un,vmesh,basis,jacf,qf,fun_set,param)
% ERRORS_FEM_2d Computes errors for 2d b.v.p.
% [errors]=errors_fem_2d(un,geom,vmesh,basis.jacf,qf,fun_set,param)
%
%        un= numerical solution
%        vmesh = structure with the mesh
%              .femtype  =2 P
%              .p = local polynomial degree (1 or 2)
%              .nodes (nodes of the mesh)
%              .nov (connectivity matrix of the mesh)
%              .ne (number of elements)
%        basis = structure with the derivatives of the Lagrange basis
%                functions (generated by calling
%                [basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree,qtype) 
%        jacf = structure with the jacobian of the maps(generated by calling
%                [basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree,qtype) 
%        qf = structure with the jacobian at the quadrature nodes (generated by calling
%                [basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree,qtype) 
%        fun_set = structure with fileds
%               .uex@(x,y) function handle with the exact solution
%               .uexx@(x,y) function handle with the d/dx of the exact solution
%               .uexy@(x,y) function handle with the d/dy of the exact solution
%        param: column array with at least these entries (the other are not
%        referenced):
%        param(8) = 0: absolute errors are computed
%                   1: relative errors are computed
%
% Output: 
%         errors: structure with fields
%           .inf = ||u_ex-u_h| with respect to discrete maximum-norm
%           .h1 = ||u_ex-u_h|| with respect to discrete H1-norm
%           .l2 = ||u_ex-u_h|| with respect to discrete L2-norm
%

%   Written by Paola Gervasio
%   $Date: 2015/04/15$


errtype=param(8);


% compute errors on the exact solution

np=length(un);

% Evaluate exact solution

U=zeros(np,1)+fun_set.uex(vmesh.nodes(:,1),vmesh.nodes(:,2));

%  compute difference

err=U-un;

% ||u_ex-u_N|| with respect to discrete maximum-norm

errors.inf=norm(err,inf);
if errtype~=0
    nuex=norm(U,inf);
    if nuex>0
    errors.inf=errors.inf/norm(U,inf);
    end
        
end

% ||u_ex-u_h|| with respect to H1 norm



% 
% 
if vmesh.p==1
[errors.h1,errors.l2]=normh1_p1_2d(errtype,un,vmesh,basis,jacf,qf,fun_set);
 
elseif vmesh.p==2
[errors.h1,errors.l2]=normh1_p2_2d(errtype,un,vmesh,basis,jacf,qf,fun_set);
end
  


end
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [err_h1,err_l2]=normh1_p1_2d(errtype,un,vmesh,basis,jacf,qf,fun_set)
%  NORMH1_P1_2d: compute the H1-norm and L2-norm errors
% for Lagrange P1 fem in 2d domains 
%  [err_h1,err_l2]=normh1_p1_2d(un,vmesh,basis,jacf,qf,fun_set)
% Input : 
%         errtype = 0 for absolute error ||u-u_ex||
%                   1 for relative error ||u-u_ex||/||u_ex||
%         un = numerical solution
%         vmesh = structure with the mesh
%              .nodes (nodes of the mesh)
%              .nov (connectivity matrix of the mesh)
%         basis = structure (derivative of the basis functions)
%         jacf = straucture with jacobians at the LGL nodes
%         qf = structure qith jacobians at the quadrature nodes
%         fun_set = stracture with exact solution
%    
% Output: err_h1 =  H1-norm error  
%         err_l2 =  L2-norm error  
%

%   Written by Paola Gervasio
%   $Date: 2018/06/29$

[~,ne]=size(vmesh.nov);

nor_l2_n=0; nor_l2_d=0; nor_h1_n=0; nor_h1_d=0;
X=[qf.xq';qf.yq'];
nq=length(qf.wq);
for ie=1:ne
    % rimappo i nodi di quadratura sull'elemento ie
novl=vmesh.nov(:,ie);
x=vmesh.nodes(novl,1); 
y=vmesh.nodes(novl,2);

jac=jacf{ie}.jac;
B=jacf{ie}.J;
B1t=jacf{ie}.Jmt;

X1=B*X+[x(1)*ones(1,nq);y(1)*ones(1,nq)];
xq=X1(1,:)';
yq=X1(2,:)';

% valuto la soluzione numerica nei nodi di quadratura
uu=un(novl);
uh_quad=basis.phi*uu;

% valuta la soluzione esatta nei nodi di quadratura
uex_quad=fun_set.uex(xq,yq);

% norma L^2 di (u-u_h)

s0=sum((uex_quad-uh_quad).^2.*qf.wq)*jac;
s0d=sum(uex_quad.^2.*qf.wq)*jac;
nor_l2_n=nor_l2_n+s0;
nor_l2_d=nor_l2_d+s0d;

% norma H^1 di (u-u_h)

% valuto il gradiente di u_h (per i P1 e' un vettore costante sul
% triangolo, quindi indip dal nodo di quadratura)
% grad phi_i=B^{-T} \hat grad (\hat phi_i)
% Per i P1, il gradiente sul singolo elemento e' un vettore costante
grad=[-1,1,0;
      -1,0,1]; 
gphi=B1t*grad;
grad_uh=gphi*uu;
% valuta il grad della soluzione esatta nei nodi di quadratura
uexx_quad=fun_set.uexx(xq,yq);
uexy_quad=fun_set.uexy(xq,yq);

sx=sum((uexx_quad-grad_uh(1)).^2.*qf.wq)*jac;
sxd=sum(uexx_quad.^2.*qf.wq)*jac;
sy=sum((uexy_quad-grad_uh(2)).^2.*qf.wq)*jac;
syd=sum(uexy_quad.^2.*qf.wq)*jac;
nor_h1_n=nor_h1_n+(s0+sx+sy);
nor_h1_d=nor_h1_d+(s0d+sxd+syd);
end

if errtype==0 % absolute error
    err_h1=sqrt(nor_h1_n);
    err_l2=sqrt(nor_l2_n);
elseif errtype==1 % relative error
    if abs(nor_h1_d)>1.d-14; err_h1=sqrt(nor_h1_n/nor_h1_d); 
    else
    err_h1=sqrt(nor_h1_n); 
    end
    if abs(nor_l2_d)>1.d-14; err_l2=sqrt(nor_l2_n/nor_l2_d); 
    else
    err_l2=sqrt(nor_l2_n); 
    end
end

end
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [err_h1,err_l2]=normh1_p2_2d(errtype,un,vmesh,basis,jacf,qf,fun_set)
%  NORMH1_P2_2d: compute the H1-norm and L2-norm error
% for Lagrange P2 fem in 2d domains 
%  [err_h1,err_l2]=normh1_p2_2d(un,vmesh,basis,jacf,qf,fun_set)
% Input : 
%         errtype = 0 for absolute error ||u-u_ex||
%                   1 for relative error ||u-u_ex||/||u_ex||
%         un = numerical solution
%         vmesh = structure with the mesh
%         basis = structure (derivative of the basis functions)
%         jacf = straucture with jacobians at the LGL nodes
%         qf = structure qith jacobians at the quadrature nodes
%         fun_set = stracture with exact solution
%    
% Output: err_h1 =  H1-norm error  
%         err_l2 =  L2-norm error  
%

%   Written by Paola Gervasio
%   $Date: 2018/06/29$

[~,ne]=size(vmesh.nov);

nor_l2_n=0; nor_l2_d=0; nor_h1_n=0; nor_h1_d=0;
X=[qf.xq';qf.yq'];
nq=length(qf.wq);
for ie=1:ne
    % rimappo i nodi di quadratura sull'elemento ie
novl=vmesh.nov(:,ie);
x=vmesh.nodes(novl,1); 
y=vmesh.nodes(novl,2);


jac=jacf{ie}.jac;
B=jacf{ie}.J;
B1t=jacf{ie}.Jmt;

X1=B*X+[x(1)*ones(1,nq);y(1)*ones(1,nq)];
xq=X1(1,:)';
yq=X1(2,:)';


% valuto la soluzione numerica nei nodi di quadratura
uu=un(novl);
uh_quad=basis.phi*uu;

% valuta la soluzione esatta nei nodi di quadratura
uex_quad=fun_set.uex(xq,yq);

% norma L^2 di (u-u_h)

s0=sum((uex_quad-uh_quad).^2.*qf.wq)*jac;
s0d=sum(uex_quad.^2.*qf.wq)*jac;
nor_l2_n=nor_l2_n+s0;
nor_l2_d=nor_l2_d+s0d;

% norma H^1 di (u-u_h)

% valuto il gradiente di u_h 
% grad phi_i=B^{-T} \hat grad (\hat phi_i)
uhx=zeros(nq,1);
uhy=zeros(nq,1);

for q=1:nq
gphi1=B1t*[basis.phix(q,:);basis.phiy(q,:)]*uu;     
uhx(q)=gphi1(1);
uhy(q)=gphi1(2);
end

% valuta il grad della soluzione esatta nei nodi di quadratura
uexx_quad=fun_set.uexx(xq,yq);
uexy_quad=fun_set.uexy(xq,yq);
sx=sum((uexx_quad-uhx).^2.*qf.wq)*jac;
sxd=sum(uexx_quad.^2.*qf.wq)*jac;
sy=sum((uexy_quad-uhy).^2.*qf.wq)*jac;
syd=sum(uexy_quad.^2.*qf.wq)*jac;
nor_h1_n=nor_h1_n+(s0+sx+sy);
nor_h1_d=nor_h1_d+(s0d+sxd+syd);

end

if errtype==0 % absolute error
    err_h1=sqrt(nor_h1_n);
    err_l2=sqrt(nor_l2_n);
elseif errtype==1 % relative error
    if abs(nor_h1_d)>1.d-14; err_h1=sqrt(nor_h1_n/nor_h1_d); 
    else
    err_h1=sqrt(nor_h1_n); 
    end
    if abs(nor_l2_d)>1.d-14; err_l2=sqrt(nor_l2_n/nor_l2_d); 
    else
    err_l2=sqrt(nor_l2_n); 
    end
end

end
